
/*--
  mbidf.c
  Copyright (C) 1996-2005 G. Manzini & G. M. Del Corso.
  All rights reserved.

  This file contains an implementation of the algorithm MB-ID described in
   "Finding Exact Solutions to the Bandwidth Minimization Problem"
  by G. Del Corso (delcorso@di.unipi.it) and G. Manzini (manzini@unipmn.it)

  This version reads the input 0/1 matrix from a file whose name is given 
  on the command line.

  This software may be used freely for any purpose. However, when distributed,
  the original source must be clearly stated, and, when the source code is
  distributed, the copyright notice must be retained and any alterations in
  the code must be clearly marked. No warranty is given regarding the quality
  of this software.
--*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <sys/times.h>
#include <limits.h>


/* ==================================================================
   from now on all comments are in italian, sorry!
   ================================================================== */

/* ==================================================================
   Il problema ha una simmetria naturale nel senso che ogni ordinamento
   puo' essere ribaltato dando luogo alla stessa banda. Per questo
   motivo si considerano solo gli ordinamenti in cui il primo nodo
   ha indice minore del secondo. L'indice del primo nodo va quindi
   da 0 a matrix_size-2, mentre nella generazione dei nuovi nodi
   si verifica che sia sempre possibile concludere con un nodo
   di indice maggiore del primo.
   ================================================================== */

/* ==================================================================
   costanti fondamentali dell'algoritmo e tipi di dati
   ================================================================== */
#define use_time_lim 0           /* !=0 to put a time limit */
#define time_limit 120           /* time limit in seconds */
#define verbose 0     /* !=0 se si vuole un output completo */
#define stampa_mat 0  /* !=0 se si vuole la stampa della matrice */
#define stampa_queue 0/* !=0 se si vuole la stampa dei nodi da piazzare */
#define maxdim 256    /* dimensione massima delle matrici */

typedef short int nodo; /* tipo usato per rappresentare un nodo */

typedef short int npassi; /* tipo usato per rappresentare entro quanti passi */
		          /* un certo nodo deve essere piazzato */

/* ===================================================================
   il tipo list deve permettere di eseguire le operazioni insiemistiche
   (membership, set difference, etc) in modo veloce. Attualmente si usa
   la rappresentazione lista di bit
   =================================================================== */
#define lsize (31+maxdim)/32 /* long necessari per maxdim elementi */
typedef struct {
  unsigned long bits[lsize];
} list;                      /* tipo usato per rappresentare una lista */
                             /* di nodi in forma compatta */

/* =====================================================================
   della matrice a noi interessa solo il grafo di adiacenza, e per questo
   motivo in matco[i] si scrive la lista di adiacenza del nodo i.
   In matex[][] e' rappresentata esplicitamente la posizione degli elementi
   nonzero.
   ===================================================================== */
list matco[maxdim];  /* rappresentazione della matrice in forma compatta */

char matex[maxdim][maxdim];  /* rappresentazione 0-1 della matrice */

int best_ordering[maxdim];   /* ordinamento trovato dall'algoritmo */
int matrix_size;             /* dimensione della matrice */
int band;                    /* ampiezza di banda da realizzare */

list lastnode_list;          /* lista dei candidati a essere ultimo nodo */

/* ===================================================================
   struttura rappresentante lo status corrente della ricerca
   =================================================================== */
typedef struct {
  nodo queue[maxdim];    /* elenco dei nodi da piazzare */
  int qsize;             /* numero dei nodi ancora da piazzare */
  int busy;              /* numero di nodi busy */
  list notfree;          /* lista compatta dei nodi non free */
  npassi expire[maxdim]; /* lista delle date di scadenza */
} status;

/* ===================================================================
   struttura contenente il risultato di un test
   =================================================================== */
typedef struct {
  double nodi_gen;       /* numero di nodi generati */
  double elapsed_time;   /* tempo trascorso */
  double elapsed_time2;  /* tempo trascorso (secondo metodo) */
  int band;              /* banda ottenuta */
} risultato;

/* ====== variabile contenente il numero di "nodi" generati ====== */
double ngen;


/* ====== definisce max se necessario ===== */
#ifndef max
#define max(A,B) ((A)>(B) ? (A) : (B))
#endif

void print_ordering(int *);


int main(int argc, char *argv[])
{
  void minimize_band(void);
  int read_matrix(char *);
  int check_matex_symm(void);

  if(argc!=2) {
    printf("USAGE:   %s matrixfile\n\n",argv[0]);
    exit(1);
  }
  matrix_size = read_matrix(argv[1]);
  // ------ check input ----------------
  if(matrix_size==0 || matrix_size>maxdim) {
    fprintf(stderr,"Illegal matrix file. Recall maxdim=%d\n", maxdim);
    exit(1);
  }
  if(check_matex_symm()==0) {
    fprintf(stderr,"Matrix must be symmetric!\n");
    exit(1);
  }  

  /* ----------------- stampa i dati del problema ----------------- */
  #if verbose
  printf("\n**********************************************************");
  printf("\n                 MB-ID   Ver 1.2\n\n");
  printf("Created on %s at %s from %s\n",__DATE__,__TIME__,__FILE__);
  printf("Parameters:\n  maxdim: %d\n",maxdim);
  #if use_time_lim
  printf("  time limit for each istance: %d secs.\n",time_limit);
  #else
  printf("  no time limit.\n");
  #endif
  printf("**********************************************************\n\n");
  #endif

  minimize_band();
  return 0;
}

/* ******************************************************************
   questa funzione esegue num test nel seguente modo: si costruisce
   una matrice in cui gli elementi entro la band maxband hanno
   probabilita' p di essere non zeri. la matrice viene poi permutata
   in maniera random e la banda minimizzata usando bida()
   ****************************************************************** */
void minimize_band(void)
{
  void build_matco(void);
  void stampa_matex(void);
  void ida(risultato *);
  risultato ris;

  build_matco();
  #if stampa_mat
  stampa_matex();
  #endif
  /* -------------- calcola la banda minima ----------------- */
  ida(&ris);
  printf("Minimum band: %d\n",ris.band);
  printf("Node gen: %10.0lf  Time: %10.4lf(u)",ris.nodi_gen,ris.elapsed_time);
  printf("  %10.4lf(u+s)\n",ris.elapsed_time2);

}

/* ***************************************************************
   strategia di minimizzazione della banda basata su ida
   *************************************************************** */
void ida(risultato *ris)
{
  int set_size(list);
  int reorder(int);
  int get_band(int *);
  /* clock_t clock(void);   restituisce microsecondi: non piu' usato */
  clock_t times(struct tms *buffer);
  struct tms r;
  clock_t start,end;
  clock_t start2,end2;
  int newb,banda,i;

  ngen=0;
  times(&r);
  start2 = (r.tms_utime+r.tms_stime);     /* user + system */
  start = r.tms_utime;                    /* user only  */

  /* ------------- calcola una banda iniziale ------- */
  banda=0;
  for(i=0;i<matrix_size;i++)
    banda=max(banda,(1+set_size(matco[i]))/2);

  /* --------- cerca aumentando la banda -------- */
  for( ; ; banda++) {
    times(&r);
    end2 = (r.tms_utime+r.tms_stime);
    #if use_time_lim
    if((end2-start2)/CLK_TCK >= time_limit) {
      printf("Time limit exceeded! -ida-\n");
      exit(1);
    }
    #endif
    
    #if verbose
    printf("band = %3d, nodes = %.0f", banda, ngen);
    printf("   CPU time %f seconds.\n",((double) (end2-start2))/CLK_TCK);
    #endif

    if(reorder(banda))
      break;
  }
  times(&r);
  end2 = (r.tms_utime+r.tms_stime);
  end = r.tms_utime; 

  printf("Optimal ordering:\n");
  print_ordering(best_ordering);
  newb = get_band(best_ordering);
  if(banda!=newb) {
    printf("Band verification failed! (%d vs %d) -ida-\n",banda, newb);
    exit(1);
  }

  /* -------- restituisce il risultato del testo ---------- */
  ris->elapsed_time =  ((double) (end-start))/CLK_TCK;
  ris->elapsed_time2 = ((double) (end2-start2))/CLK_TCK;
  ris->nodi_gen = ngen;
  ris->band = banda;

}


/* ****************************************************************
   questa routine cerca un ordinamento con ampiezza di banda d.
   La simmetria del problema e' sfruttata provando "soltanto"
   matrix_dim-1 indici nella posizione 1, e mediante la list lastnode_list.
   **************************************************************** */
int reorder(int b)
{
  list create_lastnode_list(nodo n);
  int try(status *, nodo, list,int);
  list set_empty(void);
  int set_size(list);
  status start;
  int i,j, lastnode_size;

  /* ----- inizializza la variabile globale band e lo stato iniziale ---- */
  band = b;
  for(j=0;j<matrix_size;j++) {
    start.queue[j] = j;
    start.expire[j] = band;
  }
  start.qsize = matrix_size;
  start.busy = 0;
  start.notfree = set_empty();
  /* ----------------- considera un nodo alla volta ------------ */
  for(i=0;i<matrix_size-1;i++) {  /* prende i come primo nodo */
    ngen++;
    if(set_size(matco[i])>band)
      continue;
    lastnode_list = create_lastnode_list(i);
    lastnode_size = set_size(lastnode_list);
    if(lastnode_size==0)
      continue;
    if(try(&start,i,matco[i],lastnode_size))
      return 1;
  }
  return 0;
}


/* ***********************************************************************
   Questa e' la routine principale che cerca di espandere una soluzione
   livello = numero di nodi gia' fissati
   n = nodo scelto per essere messo in posizione 'livello'
   old->queue[] = lista dei nodi non ancora piazzati (e' compreso anche n)
   old->busy = numero di nodi busy (cioe' adiacenti a nodi gia' scelti)
               prima di piazzare n
   old->notfree = lista dei nodi non free prima di piazzare n
   old->expire[] = lista delle "date di scadenza" prima di piazzare n
   toadd = lista dei nodi adicenti a n che non sono gia' busy o piazzati
   La funzione restituisce 0 se il piazzamento di n non ha continuazioni
   valide, 1 altrimenti. Se restituisce 1, n e' messo in best_ordering[]
   Si usa il fatto che i nodi in old->queue[] sono ordinati per "data di
   scadenza" crescente.
   NOTA i valori di old-> non devono essere modificati
   *********************************************************************** */
int try(status *old, nodo n, list toadd, int lastnode_num)
{
  void create_newqueue(status *, status *, nodo, list);
  int init_status(status *old, status *new, nodo n, list toadd);
  int try(status *, nodo, list, int);
  list set_difference(list, list);
  list set_union(list,list);
  list set_add(nodo, list);
  int set_size(list);
  int set_member(nodo, list);
  status new;               /* variabile contenente il prossimo status */
  int maxn,i;
  nodo nextnode;
  list adjlist;
  int newbusy, next_lastnode_num;


#if stampa_queue
  printf("(%d) [",n);
  for(i=0;i<old->qsize;i++) printf("%d ",old->queue[i]);
  printf("]\n");
#endif

  /* ------ inizializza n.qsize e verifica se n e' l'ultimo nodo ------ */
  new.qsize = old->qsize-1;
  if(new.qsize==0) {
    best_ordering[matrix_size - new.qsize -1] = n;
    return 1;
  }
  // print_ordering(best_ordering);

  /* -------------- inizializza il nuovo status ------------ */
  maxn = init_status(old, &new, n, toadd);

  /* -------------- cerca un successore --------------------- */
  for(i=0;i<maxn;i++) {
    ngen++;                          /* nuova posizione generata */
    nextnode = new.queue[i];         /* possibile successore */
    /* ------------- aggiorna lastnode_num -------------- */
    if(set_member(nextnode,lastnode_list)) {
      next_lastnode_num = lastnode_num -1;
      if( (next_lastnode_num==0)&&(new.qsize>1))
        continue;
    }
    else
      next_lastnode_num = lastnode_num;
    /* ---------- esegue verifica di compatibilita' ----------- */
    adjlist = matco[nextnode];
    toadd = set_difference(adjlist, new.notfree);
    newbusy = new.busy + set_size(toadd) - (i<new.busy ? 1 : 0);
    if(newbusy<=band) {
      /* -------- passa al livello successivo --------*/
      if(try(&new, nextnode, toadd,next_lastnode_num)) {
        best_ordering[matrix_size - new.qsize-1]=n;
        return 1;                  /* soluzione trovata */
      }
    }
  }      /* end for */
  /* ------------ nessuna continuazione trovata --------------- */
  return 0;
}


/* *************************************************************
   inizializza il nuovo status partendo dalla descrizione del
   vecchio e dal nodo che e' stato scelto. restituisce il numero di
   nodi in new->queue[] da considerare
   ************************************************************* */
int init_status(status *old, status *new, nodo n, list toadd)
{
  void create_newqueue(status *, status *, nodo, list);
  int get_max_index(status *);
  list set_union(list,list);
  list set_add(nodo, list);
  int i,j, maxn;

  /* -------------- inizializza la lista queue[] ------------ */
  create_newqueue(old, new, n, toadd);

  /* --------------- inizializza la lista dei nodi non liberi ----- */
  new->notfree = set_union(old->notfree,toadd);
  new->notfree = set_add(n,new->notfree);

  /* -------------- inizializza la lista exp_time[] --------- */
  for(i=0;i<new->busy;i++)  {
    j = new->queue[i];
    new->expire[j] = old->expire[j]-1;    /* avvicina scadenza nodi busy */
  }
  for( ;i<new->qsize;i++)
    new->expire[new->queue[i]] = band; /* i nodi free hanno scandenza = band */

  /* -------- calcola il numero di nodi da considerare -------- */
  maxn = get_max_index(new);
  if(maxn<0)
    maxn = new->qsize;   /* considera tutti i nodi in queue[] */

  return maxn;
}

/* *********************************************************************
   Questa funzione prende in input uno status (lista dei nodi busy e le
   loro date di scadenza) e calcola quanti nodi busy vanno considerati.
   La funzione restituisce -1 se vanno considerati anche i nodi non busy.
   Nota: una data di scadenza 0 indica che il nodo deve essere scelto a
   questo passo, 1 == entro il passo successivo, etc. I nodi non busy
   hanno data di scadenza==band in quanto nessun nodo adiacente e' stato
   piazzato.
   ********************************************************************** */
int get_max_index(status *n)
{
  int livello;
  int i, contatore;
  int next_time;

#if 0
  /* !!!!!!! stampa nodi da considerare !!!!!! */
  for(i=0;i<n->busy;i++)
    printf("[%d %d] ",n->queue[i],n->expire[n->queue[i]]);
  printf("\n");
  /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
#endif

  livello = 0;         /* livello iniziale */
  contatore = 1;       /* massimo numero di nodi a questo livello */
  for(i=0;i<n->busy;i++)  {
    next_time = n->expire[n->queue[i]];
    if(next_time==livello) {
      contatore--;           /* trovato nuovo nodo a questo livello */
    }
    else  {                  /* inizia nuovo livello */
      if(contatore==0)
        return i;            /* considerare solo i primi i nodi */
      else if(contatore>0) {
        contatore += next_time-livello-1;/* num. nodi per questo livello */
        livello = next_time;             /* aggiorna il livello */
      }
      else {
        printf("Scadenze inconsistenti -get_max_index-\n");
        exit(1);             /* qui non ci dovrebbe mai arrivare */
      }
    }
  }
  if(contatore<0) {
    printf("Scadenze inconsistenti -get_max_index-\n");
    exit(1);                 /* qui non ci dovrebbe mai arrivare */
  }
  if(contatore==0)
     return n->busy;         /* considera solo i nodi busy */
  else
    return -1;               /* considera anche i non busy */
}


/* *********************************************************************
   questa funzione prende in input il vecchio stato e crea la nuova lista
   queue. Vengono inizializzate le variabili new->queue[], new->qsize,
   new->busy.
   I nuovi nodi busy sono i vecchi busy + quelli in toadd. Si usa il fatto
   che la nuova coda e' relativa al nodo n, e che n \not\in toadd.
   ********************************************************************* */
void create_newqueue(status *old, status *new, nodo n, list toadd)
{
  int set_member(nodo, list);
  int n0,n1;                   /* indici in new->queue[] */
  int i;
  nodo qi;

  new->qsize = old->qsize -1;
  n0=0;                 /* -- prima posizione da occupare in new->queue[] -- */
  n1=new->qsize-1;      /* -- ultima posizione da occupare in new->queue[] - */

	/* ------- copia i nodi precedentemente busy ---- */
  for(i=0;i<old->busy;i++)  {
    if(old->queue[i]!=n)
      new->queue[n0++] = old->queue[i];
  }

  /* ------- aggiunge i nodi non busy di old->queue[] ----------- */
  for(  ;i<old->qsize;i++)  {
    qi = old->queue[i];
    if(set_member(qi, toadd))
      new->queue[n0++] = qi;        /* mette qi tra i busy */
    else if(qi!=n)
      new->queue[n1--] = qi;        /* mette qi tra i non busy */
  }
  new->busy = n0;

#if verify
  /* -------- controlla q0 e q1 --------- */
  if(n0-n1!=1)  {
    printf("Errore nel programma -create_newqueue-\n");
  	exit(1);
  }
#endif

}

/* ****************************************************************
   questa routine, dato un nodo iniziale n, calcola la lista dei
   possibili nodi finali, con il vincolo che il nodo finale deve
   avere indice maggiore del nodo iniziale
   **************************************************************** */
list create_lastnode_list(nodo n)
{
  list set_empty(void);
  list set_add(nodo, list);
  int time[maxdim], nlist[maxdim];
  int i,j,next,cur,stored;
  list last;

  /* --------- inizializza time[] ------------ */
  for(i=0;i<matrix_size;i++)
    time[i] = matrix_size-1; /* nessun vincolo */

  /* --------- memorizza il nodo iniziale --------- */
  time[n]=0;
  nlist[0]=n;
  stored=1;

  /* --------- stima le date di scadenza --------- */
  for(next=0;next<stored;next++) {
    cur=nlist[next];
    if(time[cur]+band >= matrix_size-1)
      break;      /* non piu' bound significativi */
    for(j=0;j<matrix_size;j++)
      if(matex[cur][j]) {
        if(time[j]==matrix_size-1) {
          time[j]= time[cur]+band;
          nlist[stored++] = j;
          if(stored>matrix_size) {
            printf("Something really wrong! -create_lastnode_list-\n");
            exit(1);
          }
        }
      }
  }
  /* ------ verifica la validita' delle scadenze ----- */
  for(i=0;i<matrix_size;i++)
    if((time[i]<0)||(time[i]>=matrix_size)) {
      printf("Something is wrong! -create_lastnode_list-\n");
      exit(1);
    }
  /* ------ crea la lista dei possibili nodi per l'ultima posizione ----- */
  last = set_empty();
  for(i=n+1;i<matrix_size;i++)
    if(time[i]==matrix_size-1)
      last = set_add(i,last);

  return last;
}

/* ***********************************************************************
   esegue la conversione da formato espanso della matrice alla forma
   compatta
   *********************************************************************** */
void build_matco()
{
  list set_add(nodo, list);
  list set_empty(void);
  int i,j;

  for(i=0;i<matrix_size;i++) {
    matco[i] = set_empty();
    for(j=0;j<matrix_size;j++) {
      if(matex[i][j]!=0)
        matco[i] = set_add(j,matco[i]);
    }
  }
}

void stampa_matex()
{
  int i,j;

  for(i=0;i<matrix_size;i++) {
    for(j=0;j<matrix_size;j++) {
      printf("%d ", matex[i][j]);
    }
    printf("\n");
  }
}

// return 1 if matex is symmetric, 0 otherwise
// also kill diagonal ones
int check_matex_symm()
{
  int i,j;

  for(i=0;i<matrix_size;i++) {
    matex[i][i]=0;
    for(j=i+1;j<matrix_size;j++) {
      if(matex[i][j]!=matex[j][i])
	return 0;
    }
  }
  return 1;
}


/* ***********************************************************************
   calcola l'ampiezza di banda della matrice matex[][], quando
   righe e colonne vengono riordinate con permutazione p. p[i] e'
   la riga/col che viene messa in posizione i.
   *********************************************************************** */
int get_band(int p[])
{
  int i,j,pi,pj,d,band;

  band=0;
  for(i=0;i<matrix_size;i++) {
    pi = p[i];
    for(j=i+1;j<matrix_size;j++) {
      pj = p[j];
      if(matex[pi][pj]!=0) {
        d = j-i;
        band = max(band,d);
      }
    }
  }
  return band;
}

/* *********************************************
   stampa permutazione
   ********************************************* */
void print_ordering(int p[])
{
  int i;

  printf("{");
  for(i=0;i<matrix_size;i++)
    printf(" %d",p[i]);
  printf("}\n");
}

/* ************************************************************************
   read a 0/1 matrix from file named "infile_name" and store it
   into the global array matex[][].
   The matrix must be in Mathematica format: {{0, 1},{1, 0}}
   note that all chars different from { } 0 1 are ignores and that 00 counts
   as 0 0. 
   return the size of the matrix or 0 if there was some error. 
   We do not check that the matrix is simmetric.
   NB: The code is orrible...
   ************************************************************************ */
int read_matrix_mathematica(char *infile_name)
{
  FILE *f;
  int c,size,col,row;

  if(infile_name==NULL) return 0;
  f = fopen(infile_name,"r");
  if(f==NULL) {
    perror("Unable to open matrix file!");
    return 0;
  }

  // reach first open braces (beginning of matrix)
  do 
    if((c=getc(f))==EOF) 
      { fclose(f); return 0; }
  while(c!='{');

  size = -1;
  for(row=0; row<maxdim ;row++) {
    // reach beginning of next row or end of matrix
    do 
      if((c=getc(f))==EOF) 
	{ fclose(f); return 0; }
    while(c!='{' && c!='}');
    // if end of matrix check that matrix is square
    if(c=='}') { 
      if(row==size) {fclose(f); return size;}  
      else {fclose(f); return 0;}
    }
    // ----- row begin
    col = 0;
    do {
      if((c=getc(f))==EOF) 
	{fclose(f); return 0;}
      else if(c=='0' || c=='1') {
	matex[row][col++] = c - '0';
	if(col>maxdim)
	 {fclose(f); return 0;}
      }
    } while(c!='}');
    // ----- row end
    if(col==0) 
      {fclose(f); return 0;} // empty row: error!
    else if(size==-1)
      size = col;         // set matrix size
    else if(col!=size) 
      {fclose(f); return 0;} // new size does not match 
  }
  fclose(f);
  return 0;  // too many rows!
}

/* ************************************************************************
   read a 0/1 matrix from file named "infile_name" and store it
   into the global array matex[][].
   Matrix file must contain size and then elements (usally one row per line)
   ex.:
   4
   1 0 1 0
   0 1 0 0
   1 0 1 1
   0 0 1 1
   return the size of the matrix or 0 if there was some error. 
   We do not check that the matrix is simmetric.
*************************************************************************/
int read_matrix(char *infile_name)
{
  FILE *file;
  int k,nodes,i,j;

  if(infile_name==NULL) return 0;
    file = fopen(infile_name,"r");
    if( !file ){
        printf("File %s does not exists\n",infile_name);
        return(0);
    }

    fscanf(file,"%d",&nodes);
    if(nodes>maxdim)
      {fclose(file); return 0;}

    /* reading the matrix */
    for(i=0;i<nodes;i++)
        for(j=0;j<nodes;j++){
            fscanf(file,"%d",&k);
            if( i==j ) k=0;
            matex[i][j] = k;
        }
    fclose(file);
    return nodes;
}

/* ===============================================================
   Routine per la manipolazione degli insiemi. Dipendono dalla
   rappresentazione usata per il tipo list. Queste che seguono
   assumono che list sia una struttura contentnente un array di lsize
   unsigned long (quindi un elemento n e' individuato da un intero
   tra 0 e lsize-1 (==n/32), e un intero tra 0 e 31 (==n%32))
   =============================================================== */
list set_union(list a, list b)
{
  int i;

  for(i=lsize-1;i>=0;i--)
    a.bits[i] = a.bits[i] | b.bits[i];  /* unione word per word */

  return a;    /* unione di a e b */

}

int set_member(nodo n, list a)
{
  int q,r;

  q = n / 32;
  r = n % 32;

  if((1L<<r) & a.bits[q])   /* verifica se n\in a */
    return 1;
  else
    return 0;
}

list set_add(nodo n, list a)
{
  int q,r;

  q = n / 32;
  r = n % 32;
  a.bits[q] = a.bits[q] | (1L<<r);   /* aggiunge n ad a */
  return a;
}

list set_empty()
{
  list a;
  int i;

  for(i=lsize-1;i>=0;i--)
    a.bits[i] = 0L;             /* insieme vuoto */
  return a;
}

list set_difference(list a, list b)
{
  int j;
  unsigned long i;

  for(j=lsize-1;j>=0;j--) {
    i = a.bits[j] & b.bits[j];              /* intersezione */
    a.bits[j] = a.bits[j] ^ i;              /* xor */
  }
  return a;                 /* a\b */
}


int set_size(list a)
{
  int set_size32(unsigned long);
  unsigned long u;
  int i,sum;

  sum = 0;
  for(i=lsize-1;i>=0;i--) {
    if((u=a.bits[i])!=0)
      sum += set_size32(u);
  }
  return sum;
}


int set_size32(unsigned long a)
{
  static int first_call=1;
  static char ones[256];  /* ones[i] = numero di 1 in bin(i) */
  int i,j,a_size;

  /* ---- alla prima chiamata inizializza ones[] --- */
  if(first_call) {
    first_call=0;         /* esegue una volta sola */
    for(i=0;i<256;i++) {
      ones[i]=0;
      for(j=i;j!=0;j = j>>1)
        if(j&1) ones[i]++;
    }
  }
  /* ----- calcola il numero di 1 in a, considerando 8 bits alla volta ----- */
  a_size = 0;
  while(a!=0) {
    a_size += ones[(int) a&0xFF]; /* considera gli ultimi 8 bits */
    a = a >> 8;                   /* ATTENZIONE: funziona se a e' unsigned */
  }
  return a_size;
}

