/*--
  mbpsf.c
  Copyright (C) 1996-2005 G. Manzini & G. M. Del Corso.
  All rights reserved.

  This file contains an implementation of the algorithm MB-PS described in
   "Finding Exact Solutions to the Bandwidth Minimization Problem"
  by G. Del Corso (delcorso@di.unipi.it) and G. Manzini (manzini@unipmn.it)

  This version reads the input 0/1 matrix from a file whose name is given on 
  the command line.

  This software may be used freely for any purpose. However, when distributed,
  the original source must be clearly stated, and, when the source code is
  distributed, the copyright notice must be retained and any alterations in
  the code must be clearly marked. No warranty is given regarding the quality
  of this software.
--*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <sys/times.h>
#include <limits.h>

/* ==================================================================
   from now on all comments are in italian, sorry!
   ================================================================== */

/* ==================================================================
   Il problema ha una simmetria naturale nel senso che ogni ordinamento
   puo' essere ribaltato dando luogo alla stessa banda. Per questo
   motivo si considerano solo gli ordinamenti in cui il primo nodo
   ha indice minore del secondo. L'indice del primo nodo va quindi
   da 0 a matrix_size-2, mentre nella generazione delle posizioni
   bottom l'indice dell'ultimo nodo parte da 1. All'inizio della ricerca
   si eliminano dal fronte bottom tutte le posizioni che violano
   questa regola. Questo trucco e' usato (in maniera semplificata)
   in idaband dalla versione 1.2 in poi.
   ================================================================== */

/* ==================================================================
   costanti fondamentali dell'algoritmo e tipi di dati
   ================================================================== */
#define use_time_lim 1           /* !=0 to put a time limit */
#define time_limit 120             /* time limit in seconds */
#define verbose 1        /* !=0 se si vuole un output completo */
#define stampa_mat 0     /* !=0 se si vuole la stampa della matrice */
#define stampa_queue 0   /* !=0 se si vuole la stampa dei nodi da piazzare */
#define maxdim 230       /* dimensione massima delle matrici */
#define maxbdepth 20     /* massima profondita' fronte bottom */
#define maxbsize 400000  /* massima dimensione fronte bottom */

#ifndef max
#define max(A,B) ((A)>(B) ? (A) : (B))
#endif

typedef short int nodo;   /* tipo usato per rappresentare un nodo */

typedef short int npassi; /* tipo usato per rappresentare entro quanti passi */
                          /* un certo nodo deve o non deve essere piazzato */

/* ===================================================================
   il tipo list deve permettere di eseguire le operazioni insiemistiche
   (membership, set difference, etc) in modo veloce. Attualmente si usa
   la rappresentazione lista di bit
   =================================================================== */
#define lsize (31+maxdim)/32 /* long necessari per maxdim elementi */
typedef struct {
  unsigned long bits[lsize];
} list;                      /* tipo usato per rappresentare una lista */
                             /* di nodi in forma compatta */

/* =====================================================================
   della matrice a noi interessa solo il grafo di adiacenza, e per questo
   motivo in matco[i] si scrive la lista di adiacenza del nodo i.
   In matex[][] e' rappresentata esplicitamente la posizione degli elementi
   nonzero.
   ===================================================================== */
list matco[maxdim];  /* rappresentazione della matrice in forma compatta */

char matex[maxdim][maxdim];   /* rappresentazione 0-1 della matrice */

int cur_ordering[maxdim];     /* ordinamento considerato dall'algoritmo */
int matrix_size;              /* dimensione della matrice */
int band;                     /* ampiezza di banda da realizzare */
int bdepth;                   /* profondita' fronte bottom */
int bsize;                    /* numero nodi fronte bottom */

/* ===================================================================
   struttura rappresentante lo status corrente della ricerca
   =================================================================== */
typedef struct {
  nodo queue[maxdim];    /* elenco dei nodi da piazzare */
  int qsize;             /* numero dei nodi ancora da piazzare */
  int busy;              /* numero di nodi busy */
  list notfree;          /* lista compatta dei nodi non free */
  npassi expire[maxdim]; /* lista delle date di scadenza */
} status;

/* ===================================================================
   struttura rappresentante una posizione del fronte bottom
   =================================================================== */
typedef struct {
  npassi notbefore[maxdim];/* lista dei tempi di attesa */
  nodo   tail[maxbdepth];  /* ultimi "depth" nodi di una possibile soluzione */
} bpos;

bpos *blist[maxbsize];   /* array di puntatori alle posizioni bottom */

/* ===================================================================
   struttura contenente il risultato di un test
   =================================================================== */
typedef struct {
  double nodi_gen;       /* numero di nodi generati */
  double elapsed_time;   /* tempo trascorso */
  double elapsed_time2;  /* tempo trascorso (secondo metodo) */
  int band;              /* banda ottenuta */
} risultato;

/* ====== variabile contenente il numero di "nodi" generati ====== */
double ngen;


int main(int argc, char *argv[])
{
  void minimize_band(void);
  int read_matrix(char *);
  int check_matex_symm(void);

  if(argc!=3) {
    printf("USAGE:   %s bdepth matrixfile\n\n",argv[0]);
    exit(1);
  }
  bdepth = atoi(argv[1]);
  matrix_size = read_matrix(argv[2]);
  // ------ check input ----------------
  if(matrix_size==0 || matrix_size>maxdim) {
    fprintf(stderr,"Illegal matrix file. Recall maxdim=%d\n", maxdim);
    exit(1);
  }
  if(check_matex_symm()==0) {
    fprintf(stderr,"Matrix must be symmetric!\n");
    exit(1);
  }  
  if(bdepth>maxbdepth) {
    fprintf(stderr,"Error! Max bottom depth: %d\n", maxbdepth);
    exit(1);
  }
  if(bdepth<1) {
    fprintf(stderr,"Error! Max bottom depth must be positive\n");
    exit(1);
  }

  /* ----------------- stampa i dati del problema ----------------- */
  printf("\n**********************************************************"),
  printf("\n                 MB-PS  Ver 1.1\n\n");
  printf("Created on %s at %s from %s\n",__DATE__,__TIME__,__FILE__);
  printf("Parameters:\n  maxdim: %d  ",maxdim);
  printf("maxbdepth: %d",maxbdepth);
  #if use_time_lim
  printf("  time limit for each istance: %d secs.\n",time_limit);
  #else
  printf("  no time limit.\n");
  #endif
  printf("**********************************************************\n\n");
  printf("**** Size=%d  bdepth=%d ****\n\n",matrix_size,bdepth);

  minimize_band();
  return 0;
}

/* ******************************************************************
   questa funzione esegue num test nel seguente modo: si costruisce
   una matrice in cui gli elementi entro la band maxband hanno
   probabilita' p di essere non zeri. la matrice viene poi permutata
   in maniera random e la banda minimizzata usando bida()
   ****************************************************************** */
void minimize_band(void)
{
  void build_matco(void);
  void stampa_matex(void);
  void bida(risultato *);
  risultato ris;

  build_matco();
  #if stampa_mat
  stampa_matex();
  #endif
  /* -------------- calcola la banda minima ----------------- */
  bida(&ris);
  printf("Minimum band: %d\n",ris.band);
  printf("Node gen: %10.0lf  Time: %10.4lf(u)",ris.nodi_gen,ris.elapsed_time);
  printf("  %10.4lf(u+s)\n",ris.elapsed_time2);

}


/* ***************************************************************
   strategia di minimizzazione della banda basata su bida
   *************************************************************** */
void bida(risultato *ris)
{
  int create_bottom_front(int);
  int bida_reorder(int,int);
  int set_size(list);
  int get_band_verbose(int *);
  int get_band(int *);
  void print_ordering(int *);
  /* clock_t clock(void);   restituisce mecrosecondi: non piu' usato */           
  clock_t times(struct tms *buffer);
  struct tms r;
  clock_t start,end;
  clock_t start2,end2;
  int newb,bsize,banda,i;

  ngen=0;
  times(&r);
  start2 = (r.tms_utime+r.tms_stime);     /* user + system */
  start = r.tms_utime;                    /* user only  */

  for(i=0;i<maxbsize;i++)
    blist[i]=NULL;         /* inizializzazione blist */

  /* ------------- calcola una banda iniziale ------- */
  banda=0;
  for(i=0;i<matrix_size;i++)
    banda=max(banda,(1+set_size(matco[i]))/2);

  /* --------- cerca aumentando la banda -------- */
  for( ; ; banda++) {
    bsize = create_bottom_front(banda);
    times(&r);
    end2 = (r.tms_utime+r.tms_stime);
    #if use_time_lim
    if((end2-start2)/CLK_TCK >= time_limit) {
      printf("Time limit exceeded! -bida-\n");
      exit(1);
    }
    #endif    

    #if verbose
    printf("band = %3d, nodes = %.0f", banda, ngen);
    printf("   CPU time %f seconds.",((double) (end2-start2))/CLK_TCK);
    printf(" [bsize=%d]\n",bsize);
    #endif
    if(bsize==0) continue;
    if(bida_reorder(banda,bsize))
      break;
  }
  times(&r);
  end2 = (r.tms_utime+r.tms_stime); 
  end = r.tms_utime; 

  printf("Optimal ordering:\n");
  print_ordering(cur_ordering);
  newb = get_band(cur_ordering);
  if(banda!=newb) {
    printf("Band verification failed! (%d vs %d) -bida-\n",banda, newb);
    exit(1);
  }
  /* -------- restituisce il risultato del test ---------- */
  ris->elapsed_time =  ((double) (end-start))/CLK_TCK;
  ris->elapsed_time2 = ((double) (end2-start2))/CLK_TCK;
  ris->nodi_gen = ngen;
  ris->band = banda;
}


/* ****************************************************************
   questa routine cerca un ordinamento con ampiezza di banda b.
   La simmetria del problema e' sfruttata provando "soltanto"
   matrix_dim-1, indici nella posizione (1,1) e utilizzando la procedura
   firstlast_cut() per ridurre il fronte bottom.
   **************************************************************** */
int bida_reorder(int b, int bsize)
{
  int bida_try(status *, nodo, list, int);
  int time_cut(int i, int t, int size);
  int firstlast_cut(int i, int size);
  list set_empty(void);
  int set_size(list);
  status start;
  int i,j, newsize;

  /* ----- inizializza la variabile globale band e lo stato iniziale ---- */
  band = b;
  for(j=0;j<matrix_size;j++) {
    start.queue[j] = j;
    start.expire[j] = band;
  }
  start.qsize = matrix_size;
  start.busy = 0;
  start.notfree = set_empty();
  /* ----------------- considera un nodo alla volta ------------ */
  for(i=0;i<matrix_size-1;i++) {  /* prende i come primo nodo */
    ngen++;
    if(set_size(matco[i])>band)
      continue;
    newsize = firstlast_cut(i, bsize);
    newsize = time_cut(i,0,newsize);
    if(newsize>0)
      if(bida_try(&start,i,matco[i],newsize))
        return 1;
  }
  return 0;
}

/* ***********************************************************************
   Questa e' la routine principale che cerca di espandere una soluzione
   time = numero di nodi gia' fissati
   n = nodo scelto per essere messo in posizione 'time'
   old->queue[] = lista dei nodi non ancora piazzati (e' compreso anche n)
   old->busy = numero di nodi busy (cioe' adiacenti a nodi gia' scelti)
               prima di piazzare n
   old->notfree = lista dei nodi non free prima di piazzare n
   old->expire[] = lista delle "date di scadenza" prima di piazzare n
   toadd = lista dei nodi adicenti a n che non sono gia' busy o piazzati
   La funzione restituisce 0 se il piazzamento di n non ha continuazioni
   valide, 1 altrimenti.
   Si usa il fatto che i nodi in old->queue[] sono ordinati per "data di
   scadenza" crescente.
   NOTA i valori di old-> non devono essere modificati
   *********************************************************************** */
int bida_try(status *old, nodo n, list toadd, int size)
{
  int get_max_index(status *);
  int bida_try(status *, nodo, list, int);
  int init_status(status *, status *, nodo, list);
  int time_cut(int i, int t, int size);
  list set_difference(list, list);
  int set_size(list);
  void print_bpos(bpos *);
  status new;               /* variabile contenente il prossimo status */
  int time, maxn,i;
  int newbusy,newsize;
  nodo nextnode;
  list adjlist;

#if stampa_queue
  printf("(%d) [",n);
  for(i=0;i<old->qsize;i++)
    printf("%d ",old->queue[i]);
  printf("]\n");
#endif

  time = matrix_size - old->qsize;
  cur_ordering[time++]=n;

  /* -------- controlla se la ricerca e' terminanta -------- */
  if(time+bdepth==matrix_size) {
    for(i=0;i<bdepth;i++)           /* completa la soluzione */
      cur_ordering[time+i] = blist[0]->tail[i];
    /* print_bpos(blist[0]); */
    return 1;
  }

  /* -------------- inizializza il nuovo status ------------ */
  maxn = init_status(old, &new, n, toadd);

  /* -------------- cerca un successore -------------------- */
  for(i=0;i<maxn;i++) {
    ngen++;                          /* nuova posizione generata */
    nextnode = new.queue[i];         /* possibile successore */
    adjlist = matco[nextnode];
    toadd = set_difference(adjlist, new.notfree);
    newbusy = new.busy + set_size(toadd) - (i<new.busy ? 1 : 0);
    if(newbusy<=band) {
      /* -------- ripulisce il fronte bottom -------- */
      newsize = time_cut(nextnode,time,size);
      if(newsize>0) {
        /* -------- passa al livello successivo --------*/
        if(bida_try(&new, nextnode, toadd, newsize)) {
          return 1;                  /* soluzione trovata */
        }
      }
    }
  }      /* end for */
  /* ------------ nessuna continuazione trovata --------------- */
  return 0;
}


/* *************************************************************
   inizializza il nuovo status partendo dalla descrizione del
   vecchio e dal nodo che e' stato scelto. restituisce il numero di
   nodi in new->queue[] da considerare
   ************************************************************* */
int init_status(status *old, status *new, nodo n, list toadd)
{
  void create_newqueue(status *, status *, nodo, list);
  int get_max_index(status *);
  list set_union(list,list);
  list set_add(nodo, list);
  int i,j, maxn;

  /* -------------- inizializza la lista queue[] ------------ */
  create_newqueue(old, new, n, toadd);

  /* --------------- inizializza la lista dei nodi non liberi ----- */
  new->notfree = set_union(old->notfree,toadd);
  new->notfree = set_add(n,new->notfree);

  /* -------------- inizializza la lista exp_time[] --------- */
  for(i=0;i<new->busy;i++)  {
    j = new->queue[i];
    new->expire[j] = old->expire[j]-1;    /* avvicina scadenza nodi busy */
  }
  for( ;i<new->qsize;i++)
    new->expire[new->queue[i]] = band; /* i nodi free hanno scandenza = band */

  /* -------- calcola il numero di nodi da considerare -------- */
  maxn = get_max_index(new);
  if(maxn<0)
    maxn = new->qsize;   /* considera tutti i nodi in queue[] */

  return maxn;
}

/* *********************************************************************
   Questa funzione prende in input uno status (lista dei nodi busy e le
   loro date di scadenza) e calcola quanti nodi busy vanno considerati.
   La funzione restituisce -1 se vanno considerati anche i nodi non busy.
   Nota: una data di scadenza 0 indica che il nodo deve essere scelto a
   questo passo, 1 == entro il passo successivo, etc. I nodi non busy
   hanno data di scadenza==band in quanto nessun nodo adiacente e' stato
   piazzato.
   ********************************************************************** */
int get_max_index(status *n)
{
  int livello;
  int i, contatore;
  int next_time;

#if 0
  /* !!!!!!! stampa nodi da considerare !!!!!! */
  for(i=0;i<n->busy;i++)
    printf("[%d %d] ",n->queue[i],n->expire[n->queue[i]]);
  printf("\n");
  /* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
#endif

  livello = 0;         /* livello iniziale */
  contatore = 1;       /* massimo numero di nodi a questo livello */
  for(i=0;i<n->busy;i++)  {
    next_time = n->expire[n->queue[i]];
    if(next_time==livello) {
      contatore--;           /* trovato nuovo nodo a questo livello */
    }
    else  {                  /* inizia nuovo livello */
      if(contatore==0)
        return i;            /* considerare solo i primi i nodi */
      else if(contatore>0) {
        contatore += next_time-livello-1;/* num. nodi per questo livello */
        livello = next_time;             /* aggiorna il livello */
      }
      else {
        printf("Scadenze inconsistenti -get_max_index-\n");
        exit(1);             /* qui non ci dovrebbe mai arrivare */
      }
    }
  }
  if(contatore<0) {
    printf("Scadenze inconsistenti -get_max_index-\n");
    exit(1);                 /* qui non ci dovrebbe mai arrivare */
  }
  if(contatore==0)
     return n->busy;         /* considera solo i nodi busy */
  else
    return -1;               /* considera anche i non busy */
}


/* *********************************************************************
   questa funzione prende in input il vecchio stato e crea la nuova lista
   queue. Vengono inizializzate le variabili new->queue[], new->qsize,
   new->busy.
   I nuovi nodi busy sono i vecchi busy + quelli in toadd. Si usa il fatto
   che la nuova coda e' relativa al nodo n, e che n \not\in toadd.
   ********************************************************************* */
void create_newqueue(status *old, status *new, nodo n, list toadd)
{
  int set_member(nodo, list);
  int n0,n1;                   /* indici in new->queue[] */
  int i;
  nodo qi;

  new->qsize = old->qsize -1;
  n0=0;                 /* -- prima posizione da occupare in new->queue[] -- */
  n1=new->qsize-1;      /* -- ultima posizione da occupare in new->queue[] - */

	/* ------- copia i nodi precedentemente busy ---- */
  for(i=0;i<old->busy;i++)  {
    if(old->queue[i]!=n)
      new->queue[n0++] = old->queue[i];
  }

  /* ------- aggiunge i nodi non busy di old->queue[] ----------- */
  for(  ;i<old->qsize;i++)  {
    qi = old->queue[i];
    if(set_member(qi, toadd))
      new->queue[n0++] = qi;        /* mette qi tra i busy */
    else if(qi!=n)
      new->queue[n1--] = qi;        /* mette qi tra i non busy */
  }
  new->busy = n0;

#if verify
  /* -------- controlla q0 e q1 --------- */
  if(n0-n1!=1)  {
    printf("Errore nel programma -create_newqueue-\n");
  	exit(1);
  }
#endif

}



/* ==============================================================
   =                                                            =
   =  procedure per la ricerca bottom                           =
   =                                                            =
   ============================================================== */

/* ***************************************************************
   creazione e memorizzazione (in blist[]) delle posizioni del bottom
   front Tali posizioni sono quelle ottenute piazzando "bdepth" nodi
   in modo che il segmento risultante abbia banda minore di "b"
   restituisce il numero di nodi messi in blist[]
   Nota: la memorizzazione in blist[] e' effettuata dalla funzione
   bottom_explore() quando raggiunge il livello bdepth
   *************************************************************** */
int create_bottom_front(int b)
{
  void bottom_explore(status *, nodo, list,int);
  list set_empty(void);
  int set_size(list);
  status start;
  int i,j;

  /* ------ azzera l'elenco dei nodi in blist[] -------- */
  bsize=0;

  /* ----- inizializza la variabile globale band e lo stato iniziale ---- */
  band = b;
  for(j=0;j<matrix_size;j++) {
    start.queue[j] = j;
    start.expire[j] = band;
  }
  start.qsize = matrix_size;
  start.busy = 0;
  start.notfree = set_empty();
  /* -------------------------------------------------------------------
     considera un nodo alla volta come ultimo nodo della sol. ed espande.
     i>=1 perche' l'ultimo nodo lo prendiamo > del primo.
     -------------------------------------------------------------------- */
  for(i=1;i<matrix_size;i++) {    /* prende i come ultimo nodo */
    ngen++;
    if(set_size(matco[i])>band)
      continue;
    else
      bottom_explore(&start,i,matco[i],0);
  }
  return bsize;
}


/* ***********************************************************************
   Questa e' la routine principale usata nella creazione del fronte
   bottom. E' stata ottenuta modificando la procedura try() di idaband.c
   livello = numero di nodi gia' fissati
   n = nodo scelto per essere messo in posizione 'livello'
   old->queue[] = lista dei nodi non ancora piazzati (e' compreso anche n)
   old->busy = numero di nodi busy (cioe' adiacenti a nodi gia' scelti)
               prima di piazzare n
   old->notfree = lista dei nodi non free prima di piazzare n
   old->expire[] = lista delle "date di scadenza" prima di piazzare n
   toadd = lista dei nodi adiacenti a n che non sono gia' busy o piazzati
   Si usa il fatto che i nodi in old->queue[] sono ordinati per "data di
   scadenza" crescente.
   NOTA i valori di old-> non devono essere modificati
   *********************************************************************** */
void bottom_explore(status *old, nodo n, list toadd,int livello)
{
  void create_newqueue(status *, status *, nodo, list);
  int get_max_index(status *);
  void bottom_explore(status *, nodo, list,int);
  int init_status(status *, status *, nodo, list);
  list set_difference(list, list);
  int set_size(list);
  void store_bpos(nodo *);
  status new;               /* variabile contenente il prossimo status */
  int newbusy,maxn,i;
  nodo nextnode;
  list adjlist;
  static nodo bpath[maxbdepth];              /* nodi nel path corrente */

  /* -- aggiunge n a bpath e alla sua rappresentazione compatta -- */
  bpath[livello++] = n;

  /* -- se raggiunto livello bdepth, aggiunge new in blist[] -- */
  if(livello==bdepth) {
    store_bpos(bpath);
    return;
  }

  /* ----------- inizializza la lista queue[] di new ----------- */
  maxn = init_status(old, &new, n, toadd);

  /* -------------- cerca un successore --------------------- */
  for(i=0;i<maxn;i++) {
    ngen++;                          /* nuova posizione generata */
    nextnode = new.queue[i];         /* possibile successore */
    adjlist = matco[nextnode];
    toadd = set_difference(adjlist, new.notfree);
    newbusy = new.busy + set_size(toadd) - (i<new.busy ? 1 : 0);
    if(newbusy<=band) {
      /* -------- passa al livello successivo --------*/
      bottom_explore(&new, nextnode, toadd,livello);
    }
  }      /* end for */
}


/* *************************************************************
   crea una posizione del fronte bottom usando la lista di
   bdepth nodi in invtail[]
   ************************************************************* */
void store_bpos(nodo *invtail)
{
  nodo n, aux[maxdim];
  bpos *p;
  int i,j,auxmin;

  /* ------- alloca la memoria se necessario ------ */
  if(blist[bsize]==NULL) {
    blist[bsize]=(bpos *) malloc(sizeof(bpos));
    if(blist[bsize]==NULL) {
      printf("Not enough memory! (%d pos created) -store_bpos-\n",bsize);
      exit(1);
    }
  }
  /* ------ copia il puntatore da usare e incrementa bsize ------ */
  p = blist[bsize++];
  if(bsize>=maxbsize) {
    printf("Bottom front greater than %d! -store_bpos-\n",maxbsize);
    exit(1);
  }

  /* ------ copia bpath[] in blist[bsize] ------ */
  for(i=0;i<bdepth;i++)
    p->tail[bdepth-1-i] = invtail[i]; /* inverte l'ordine */

  /* ------ inizializza notbefore[] e aux[]------ */
  for(i=0;i<matrix_size;i++)
    p->notbefore[i] = 0;
  for(i=0;i<bdepth;i++) {
    aux[matrix_size-1-i] = invtail[i];
    p->notbefore[invtail[i]] = matrix_size-1-i;
  }
  auxmin=matrix_size-bdepth;

  /* --------- crea la lista dei tempi di attesa -------- */
  for(i=matrix_size-1;i>=auxmin;i--) {
    n = aux[i];
    if(p->notbefore[n]<=band)
      break;                       /* non si ottengono piu' limiti utili */
    for(j=0;j<matrix_size;j++) {   /* considera gli adiacenti ad n */
      if(matex[n][j]) {
        /* --- trovato j adiacente ad n ---- */
        if(p->notbefore[j]==0) {
          p->notbefore[j]= p->notbefore[n] - band;
	       aux[--auxmin]=j;
          if(auxmin<0) {
            printf("Something really wrong! -store_bpos-\n");
            exit(1);
          }
        }
      }
    }
  }
}

/* ***********************************************************************
   questa procedura elimina dal bottom front tutte le posizioni in cui
   l'ultimo nodo ha indice maggiore di n (si sta sfruttando la simmetria del
   problema). L'eliminazione e' fatta mettendo i punatori a nodi bad in fondo
   e restituendo il numero di nodi good
   *********************************************************************** */
int firstlast_cut(int n, int size)
{
  int good,bad;
  bpos *temp;

  good = -1;
  bad  = size;
  while(1) {
    do {
      if(++good>=size) break;
    } while(blist[good]->tail[bdepth-1]>n);
    do {
      if(--bad<0) break;
    } while(blist[bad]->tail[bdepth-1] <= n);
    if(good<bad) {
      temp = blist[good];
      blist[good] = blist[bad];
      blist[bad] = temp;
    }
    else
      break;
  }
  return good;
}

/* ***********************************************************************
   questa procedura elimina dal bottom front tutte le posizioni in cui
   il nodo n deve apparire dopo il tempo t. L'eliminazione e' fatta mettendo i
   punatori a nodi bad in fondo e restituendo il numero di nodi good
   *********************************************************************** */
int time_cut(int n, int t, int size)
{
  int good,bad;
  bpos *temp;

  good = -1;
  bad  = size;
  while(1) {
    do {
      if(++good>=size) break;
    } while(blist[good]->notbefore[n]<=t);
    do {
      if(--bad<0) break;
    } while(blist[bad]->notbefore[n] > t);
    if(good<bad) {
      temp = blist[good];
      blist[good] = blist[bad];
      blist[bad] = temp;
    }
    else
      break;
  }
  return good;
}

void print_bpos(bpos *p)
{
  int i;

  printf("[");
  for(i=0;i<bdepth;i++)
    printf("%2d ",p->tail[i]);
  printf("]\n");

  for(i=0;i<matrix_size;i++)
    printf("(%2d %2d) ",i,p->notbefore[i]);
  printf("\n");
}



/* =============================================================
   =                                                           =
   =  procedure di servizio: input, output, conversione, etc.  =
   =                                                           =
   ============================================================= */

/* ***********************************************************************
   esegue la conversione da formato espanso della matrice alla forma
   compatta
   *********************************************************************** */
void build_matco()
{
  list set_add(nodo, list);
  list set_empty(void);
  int i,j;

  for(i=0;i<matrix_size;i++) {
    matco[i] = set_empty();
    for(j=0;j<matrix_size;j++) {
      if(matex[i][j]!=0)
        matco[i] = set_add(j,matco[i]);
    }
  }
}

void stampa_matex()
{
  int i,j;

  for(i=0;i<matrix_size;i++) {
    for(j=0;j<matrix_size;j++) {
      printf("%d ", matex[i][j]);
    }
    printf("\n");
  }
}

// return 1 if matex is symmetric, 0 otherwise
// also kill diagonal ones
int check_matex_symm()
{
  int i,j;

  for(i=0;i<matrix_size;i++) {
    matex[i][i]=0;
    for(j=i+1;j<matrix_size;j++) {
      if(matex[i][j]!=matex[j][i])
	return 0;
    }
  }
  return 1;
}


/* ***********************************************************************
   calcola l'ampiezza di banda della matrice matex[][], quando
   righe e colonne vengono riordinate con permutazione p. p[i] e'
   la riga/col che viene messa in posizione i.
   *********************************************************************** */
int get_band(int p[])
{
  int i,j,pi,pj,d,band;

  band=0;
  for(i=0;i<matrix_size;i++) {
    pi = p[i];
    for(j=i+1;j<matrix_size;j++) {
      pj = p[j];
      if(matex[pi][pj]!=0) {
        d = j-i;
        band = max(band,d);
      }
    }
  }
  return band;
}

/* *********************************************
   come sopra ma stampa i dati di ogni riga
   ********************************************* */
int get_band_verbose(int p[])
{
  int i,j,pi,pj,d,band,iband,ij;

  band=0;
  for(i=0;i<matrix_size;i++) {
    pi = p[i];
    iband=0; ij=i;
    for(j=i+1;j<matrix_size;j++) {
      pj = p[j];
      if(matex[pi][pj]!=0) {
        d = j-i;
        if(d>iband) {
          iband = d;
          ij = pj;
        }
      }
    }
    printf("[%2d %2d   %2d]    ",pi,ij,iband);
    band = max(band,iband);
  }
  printf("\n");
  return band;
}

/* *********************************************
   stampa permutazione
   ********************************************* */
void print_ordering(int p[])
{
  int i;

  printf("{");
  for(i=0;i<matrix_size;i++)
    printf(" %d",p[i]);
  printf("}\n");
}

/* ************************************************************************
   read a 0/1 matrix from file named "infile_name" and store it
   into the global array matex[][].
   The matrix must be in Mathematica format: {{0, 1},{1, 0}}
   note that all chars different from { } 0 1 are ignored and that 00 counts
   as 0 0. 
   return the size of the matrix or 0 if there was some error. 
   We do not check that the matrix is simmetric.
   NB: The code is orrible...
   ************************************************************************ */
int read_matrix_mathematica(char *infile_name)
{
  FILE *f;
  int c,size,col,row;

  if(infile_name==NULL) return 0;
  f = fopen(infile_name,"r");
  if(f==NULL) {
    perror("Unable to open matrix file!");
    return 0;
  }

  // reach first open braces (beginning of matrix)
  do 
    if((c=getc(f))==EOF) 
      { fclose(f); return 0; }
  while(c!='{');

  size = -1;
  for(row=0; row<maxdim ;row++) {
    // reach beginning of next row or end of matrix
    do 
      if((c=getc(f))==EOF) 
	{ fclose(f); return 0; }
    while(c!='{' && c!='}');
    // if end of matrix check that matrix is square
    if(c=='}') { 
      if(row==size) {fclose(f); return size;}  
      else {fclose(f); return 0;}
    }
    // ----- row begin
    col = 0;
    do {
      if((c=getc(f))==EOF) 
	{fclose(f); return 0;}
      else if(c=='0' || c=='1') {
	matex[row][col++] = c - '0';
	if(col>maxdim)
	 {fclose(f); return 0;}
      }
    } while(c!='}');
    // ----- row end
    if(col==0) 
      {fclose(f); return 0;} // empty row: error!
    else if(size==-1)
      size = col;         // set matrix size
    else if(col!=size) 
      {fclose(f); return 0;} // new size does not match 
  }
  fclose(f);
  return 0;  // too many rows!
}

/* ************************************************************************
   read a 0/1 matrix from file named "infile_name" and store it
   into the global array matex[][].
   Matrix file must contain size and then elements (usally one row per line)
   ex.:
   4
   1 0 1 0
   0 1 0 0
   1 0 1 1
   0 0 1 1
   return the size of the matrix or 0 if there was some error. 
   We do not check that the matrix is simmetric.
*************************************************************************/
int read_matrix(char *infile_name)
{
  FILE *file;
  int k,nodes,i,j;

  if(infile_name==NULL) return 0;
    file = fopen(infile_name,"r");
    if( !file ){
        printf("File %s does not exists\n",infile_name);
        return(0);
    }

    fscanf(file,"%d",&nodes);
    if(nodes>maxdim)
      {fclose(file); return 0;}

    /* reading the matrix */
    for(i=0;i<nodes;i++)
        for(j=0;j<nodes;j++){
            fscanf(file,"%d",&k);
            if( i==j ) k=0;
            matex[i][j] = k;
        }
    fclose(file);
    return nodes;
}



/* ===============================================================
   Routine per la manipolazione degli insiemi. Dipendono dalla
   rappresentazione usata per il tipo list. Queste che seguono
   assumono che list sia una struttura contenente un array di lsize
   unsigned long (quindi un elemento n e' individuato da un intero
   tra 0 e lsize-1 (==n/32), e un intero tra 0 e 31 (==n%32))
   =============================================================== */
list set_union(list a, list b)
{
  int i;

  for(i=lsize-1;i>=0;i--)
    a.bits[i] = a.bits[i] | b.bits[i];  /* unione word per word */

  return a;    /* unione di a e b */

}

int set_member(nodo n, list a)
{
  int q,r;

  q = n / 32;
  r = n % 32;

  if((1L<<r) & a.bits[q])   /* verifica se n\in a */
    return 1;
  else
    return 0;
}

list set_add(nodo n, list a)
{
  int q,r;

  q = n / 32;
  r = n % 32;
  a.bits[q] = a.bits[q] | (1L<<r);   /* aggiunge n ad a */
  return a;
}

list set_empty()
{
  list a;
  int i;

  for(i=lsize-1;i>=0;i--)
    a.bits[i] = 0L;             /* insieme vuoto */
  return a;
}

list set_difference(list a, list b)
{
  int j;
  unsigned long i;

  for(j=lsize-1;j>=0;j--) {
    i = a.bits[j] & b.bits[j];              /* intersezione */
    a.bits[j] = a.bits[j] ^ i;              /* xor */
  }
  return a;                 /* a\b */
}


int set_size(list a)
{
  int set_size32(unsigned long);
  unsigned long u;
  int i,sum;

  sum = 0;
  for(i=lsize-1;i>=0;i--) {
    if((u=a.bits[i])!=0)
      sum += set_size32(u);
  }
  return sum;
}


int set_size32(unsigned long a)
{
  static int first_call=1;
  static char ones[256];  /* ones[i] = numero di 1 in bin(i) */
  int i,j,a_size;

  /* ---- alla prima chiamata inizializza ones[] --- */
  if(first_call) {
    first_call=0;         /* esegue una volta sola */
    for(i=0;i<256;i++) {
      ones[i]=0;
      for(j=i;j!=0;j = j>>1)
        if(j&1) ones[i]++;
    }
  }
  /* ----- calcola il numero di 1 in a, considerando 8 bits alla volta ----- */
  a_size = 0;
  while(a!=0) {
    a_size += ones[(int) a&0xFF]; /* considera gli ultimi 8 bits */
    a = a >> 8;                   /* ATTENZIONE: funziona se a e' unsigned */
  }
  return a_size;
}


